"""
Description
-----------

:func:`flasket.cli.make_flasket` allows a user to quickly create a Flasket application. It's an all-in-one command-line oriented function.

Functions
---------

make_flasket
^^^^^^

.. autofunction:: make_flasket


"""
import argparse
import os
import sys
import typing as t

from torxtools import xdgtools
from torxtools.cfgtools import which
from torxtools.pathtools import expandpath
from yaml import safe_load

from ._utils import deepmerge
from .defaults import CFGFILE_SEARCH_PATHS, default_configuration
from .middleware import FlaskFlasketCreator, GunicornFlasketCreator

__all__: t.List[str] = [
    "make_flasket",
]


def _read(cfgfile: str) -> t.Dict:
    """
    Convenience function in order to be mocked.

    Parameters
    ----------
    cfgfile: str

        a single path representing a yaml file.

    Returns
    -------
    dict:

        a dictionary
    """
    with open(cfgfile, encoding="UTF-8") as fd:
        data = safe_load(fd)
    return data or {}


def _parse_arguments(
    *,
    argv,
    cfgfile,
    cfgfilepaths,
    creator_class,
    defaults,
    description,
):
    """ """
    if not argv:
        return {}

    # argument_default=None does not set the default to None for boolean options,
    # so we'll specifically set default=None for those values
    #
    # Default values aren't actually added/set here, but in the FlasketSettings,
    # We only care about values that were specified.
    parser = argparse.ArgumentParser(
        description=description,
        argument_default=None,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    # Create helptext for UI option
    b_ui = {True: "enabled", False: "disabled"}[defaults["server"]["ui"]]
    if cfgfilepaths:
        # Keep on two lines, otherwise line continuation will make
        # an obsure argparse bug appear
        helpmsg_cfgfile = "Use CFGFILE as configuration file, "
        helpmsg_cfgfile += f"otherwise first file found in search path is used. (default search path: {cfgfilepaths})"
    else:
        helpmsg_cfgfile = "Use CFGFILE as configuration file."

    # fmt: off
    parser.add_argument(
        "-l", "--listen", metavar="HOST",
        help=f'The ip to listen on (default: {defaults["server"]["listen"]})',
    )
    parser.add_argument(
        "-p", "--port", metavar="PORT", type=int,
        help=f'The port to listen on (default: {defaults["server"]["port"]})',
    )
    # Is configuration file already specified?
    if not cfgfile:
        parser.add_argument(
        "-c", "--cfgfile", metavar="CFGFILE",
        help=helpmsg_cfgfile,
        )
    parser.add_argument(
        "--ui", action="store_true", default=None,
        help=f"Enable the OpenAPI UI. Disable with --no-ui. (default: {b_ui})",
    )
    parser.add_argument(
        "--no-ui", action="store_false", default=None, dest="ui",
        help=argparse.SUPPRESS,
    )
    # fmt: on
    creator_class.add_arguments(defaults["server"], parser)
    args = parser.parse_args(argv)
    return vars(args)


def make_flasket(
    rootpath: str,
    middleware: str = "gunicorn",
    *,
    argv: t.List[str] = None,
    cfgfile=None,
    cfgfilename=None,
    cfgfilepaths=None,
    default_cfg: t.Dict[str, t.Any] = None,
    description="Flasket server",
    options=None,
):
    """
    Parameters
    ----------
    rootpath: str

        path of location where the 'api', 'htdocs' and 'app' directories will be served.

    middleware: str

        Middleware to use: 'gunicorn' for production, 'flask' for debug.

    argv: list[str]

        Uses :code:`sys.argv[1:]` if None. Use :code:`[]` if you desire to deactive argument parsing.

    cfgfile: str

        Configuration file to use. Overrides searching and asking for a configuration file.

    cfgfilename: str

        Configuration filename to use. Typically this is your packagename with a 'yml' extension.

    cfgfilepaths: str

        Paths to search for in order to find the configuration file. Every value of the list must contain the placeholder '{cfgfilename}' such as :code:`["/home/jdoe/{cfgfilename}"]`

    default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

        Dictionary containing the defaults for the command line arguments and configuration.
        Passed value will be merged with the default factory configuration, command line arguments
        and the configuration file that was read.

    description: str

        Text description to use in argparse.

    options: dict

        Options to be passed down to backends.

        .. code-block:: python

          {"static": {"redirect_to": "/app/", "redirect_code": 308}}
    """
    # Verify input
    assert True in [argv is None, isinstance(argv, list)], "error: argv must be None or a list"

    # rootpath was probably passed by "__file__"
    if rootpath and os.path.isfile(rootpath):
        rootpath = os.path.dirname(rootpath)

    # Sets environment variables for XDG paths
    xdgtools.setenv()

    # Verify middleware exists
    middleware = middleware.lower().strip()
    if middleware not in ["flask", "gunicorn"]:
        raise ValueError('middleware argument must be in ["flask", "gunicorn", "test"]')
    if middleware == "flask":
        creator_class = FlaskFlasketCreator()
    if middleware == "gunicorn":
        creator_class = GunicornFlasketCreator()

    # Prepare search path for configuration file. Disable it if we're looking
    # for a impossible (None) file name
    if cfgfilepaths is None:
        cfgfilepaths = CFGFILE_SEARCH_PATHS
    cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]

    # Create a default configuration from what was passed
    # and what we set. Other values are filtered
    defaults = default_configuration(default_cfg)

    # Parse arguments if they exist
    if argv is None:
        argv = sys.argv[1:]

    arguments = _parse_arguments(
        argv=argv,
        cfgfile=cfgfile,
        cfgfilepaths=cfgfilepaths,
        creator_class=creator_class,
        defaults=defaults,
        description=description,
    )

    # Copy arguments over to an dict with server
    arguments = {k: v for k, v in arguments.items() if v is not None}
    cfgfile = arguments.pop("cfgfile", cfgfile)
    arguments = {"server": arguments}

    # Search for the configuration file
    if not cfgfile and cfgfilepaths:
        # Search for cfgfile
        cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
        cfgfile = which(cfgfile, expandpath(cfgfilepaths))

    filedata = _read(cfgfile) or {}

    # Merge in the reverse order of priority
    cfg = defaults
    cfg = deepmerge(cfg, filedata)
    cfg = deepmerge(cfg, arguments)
    creator_class.force_cfg(defaults["server"])

    return creator_class.flasket(rootpath=rootpath, cfg=cfg, options=options)



from attrs import define

from ..properties import BackendProperties


@define(kw_only=True, slots=False)
class Backend(BackendProperties):
    def __post_init__(self) -> None:
        super().__post_init__()
        # self.flasket is already a weakref
        self.flask.flasket = self.flasket
        self._copy_configuration()

    def _copy_configuration(self):
        for key, value in self.flasket.flask.config.items():
            self.flask.config[key] = value

    @property
    def flasket(self) -> "Flasket":
        """
        Return Flasket weak reference.
        """
        return self._flasket



import functools
import logging
import typing as t

import flask
from attr import define, field

from .exceptions import HTTPExceptions

__all__: t.List[str] = [
    "FlaskProperties",
    "FlasketProperties",
    "LoggerProperties",
]


@define(kw_only=True, slots=False)
class LoggerProperties:
    """
    Minimal class for inheritance of the logger property
    """

    @property
    def logger(self) -> logging.Logger:
        """
        Return a :class:`logging.Logger` named ``stderr``
        """
        return logging.getLogger("stderr")


@define(kw_only=True, slots=False)
class FlasketProperties(HTTPExceptions):
    #: Parent flasket application
    _flasket = field()
    _options = field(default=None)

    # The class accesses fields that exist only
    # on the flasket class
    @property
    def debug(self) -> bool:
        """
        Return debug mode (True/False)
        """
        return self._flasket._debug

    @property
    def config(self) -> t.Dict[str, t.Any]:
        """
        Return Flasket configuration used at start
        """
        return self._flasket._cfg

    @property
    def options(self) -> t.Dict[str, t.Any]:
        """
        Return Flasket options used at start
        """
        return self._flasket._cfg

    @property
    def rootpath(self) -> str:
        """
        Return Flasket rootpath used at start
        """
        return self._flasket._rootpath

    @property
    def clients(self) -> object:
        return self._flasket._clients

    @property
    def sitename(self) -> str:
        """
        Return combination of network host and port
        """
        return self._flasket.sitename

    @property
    def host(self) -> str:
        """
        Network hosts on which Flasket is listening on.
        """
        return self._flasket.host

    @property
    def port(self) -> int:
        """
        Network post on which Flasket is listening on.
        """
        return self._flasket.port

    @property
    def baseurl(self) -> str:
        """
        Returns baseurl "{scheme}://{netloc}"
        """
        return self._flasket.baseurl


@define(kw_only=True, slots=False)
class FlaskProperties:
    _flask = field(default=None, init=False)

    @property
    def flask(self) -> flask.Flask:
        """
        Return underlying Flask object
        """
        return self._flask

    def __call__(self, environ: dict, start_response):
        return self._flask.wsgi_app(environ, start_response)

    @property
    def current_app(self) -> t.Any:
        """
        Return Flask current_app
        """
        return flask.current_app

    @property
    def session(self) -> t.Any:
        """
        Return Flask session
        """
        return flask.session

    @property
    def request(self) -> t.Any:
        """
        Return Flask request
        """
        return flask.request

    @staticmethod
    def redirect(location, code=302) -> t.Any:
        """
        Redirect with a 302 Found
        """
        return flask.redirect(location, code=code)

    @staticmethod
    def make_response(*args) -> t.Any:
        """
        Create a Flask response object
        """
        return flask.make_response(
            *args,
        )

    @staticmethod
    def render_template(template_name_or_list, **context):
        return flask.render_template(template_name_or_list, **context)


@define(kw_only=True, slots=False)
class BackendProperties(FlasketProperties, FlaskProperties, LoggerProperties):
    def __post_init__(self) -> None:
        # Register a route to always call before/after_app_request
        self._flasket._register_fake_route(handler=self)

    # pylint: disable=bad-staticmethod-argument
    @staticmethod
    def post_init(cls):
        @functools.wraps(cls, updated=())
        class Wrapper(cls):
            def __attrs_post_init__(self):
                super().__attrs_post_init__()
                super().__post_init__()

        return Wrapper



r"""
Description
-----------

Options
-------

.. program:: flasket

.. option:: -l HOST, --listen HOST

   The ip to listen on (default: localhost)

.. option:: -p PORT, --port PORT

   The port to listen on (default: 8080)

.. option:: -c CFGFILE, --cfgfile CFGFILE

   Use CFGFILE as configuration file, otherwise first file found in
   search path is used. (default search path: (...))

.. option:: --ui, --no-ui

   Enable the OpenAPI UI. Disable with :option:`--no-ui`. (default: enabled)

.. option::  -w WORKERS, --workers WORKERS

   Number of thread workers. (default: 0. If 0, cpu to use is (cpu_count * 2) with a maximum
   of 8; if negative, cpu to use is (cpu_count * 2) with no maximum.)

.. option::  --pidfile FILE

   A filename to use for the PID file. (default: none)

Usage
-----

.. program:: flasket

.. code-block:: python

  #!/usr/bin/env python3

  from flasket import make_flasket

  if __name__ == "__main__":
      flasket = make_flasket(__file__, "gunicorn", cfgname="example.yml")
      flasket.run()

"""
import argparse
import logging
import multiprocessing
import typing as t

from attrs import define, field
from flask.logging import default_handler
from gunicorn.app.base import BaseApplication
from torxtools.pathtools import expandpath

from ..flasket import Flasket

__all__: t.List[str] = [
    "GunicornApplication",
    "GunicornFlasketCreator",
]


class GunicornFlasketCreator:
    @staticmethod
    def flasket(*args: t.Any, **kwargs: t.Any) -> Flasket:
        app = GunicornApplication(*args, **kwargs)
        return app._flasket

    @staticmethod
    def force_cfg(cfg: t.Dict["str", t.Any]) -> None:
        cfg["debug"] = False

    @staticmethod
    def add_arguments(defaults: t.Dict["str", t.Any], parser: argparse.ArgumentParser) -> None:
        """
        Adds Gunicorn middleware specific command line arguments :option:`--workers`, :option:`--pidfile`
        to the ArgumentParser.

        Parameters
        ----------
        defaults:

            a dictionnary containing default values

        parser: argparse.ArgumentParser

            a parser on which `add_argument` will be called
        """
        workers = defaults["workers"]
        pidfile = defaults["pidfile"]
        if not pidfile:
            pidfile = "none"

        # fmt: off
        parser.add_argument(
            "-w", "--workers", metavar="WORKERS", type=int,
            help=f"Number of thread workers. (default: {workers}. If 0, cpu to use is (cpu_count * 2) with a maximum of 8; if negative, cpu to use is (cpu_count * 2) with 8 maximum.)",
        )
        parser.add_argument(
            "--pidfile", metavar="FILE", type=str,
            help=f"A filename to use for the PID file. (default: {pidfile})",
        )
        # fmt: on


class GunicornApplication(BaseApplication):  # type: ignore[misc]
    # .venv/lib/python3.9/site-packages/gunicorn/app/base.py
    """
    Gunicorn WSGI application class
    """

    @staticmethod
    def _get_cpu_workers(workers: int) -> int:
        """
        Return the number of workers gunicorn should start with.
        Impose a max limit to avoid overloading the host.

        Parameters
        ----------
        workers: int

            number of workers configured

        Returns
        -------
        int
        """
        if workers > 0:
            return workers

        # Do not overload the host with a default value too large
        # if it's a container, it could report host cpus
        cpus = max((multiprocessing.cpu_count() * 2), 1)
        if workers == 0:
            return min(cpus, 8)
        return cpus

    # pylint: disable=abstract-method
    # http://docs.gunicorn.org/en/stable/custom.html
    def __init__(
        self,
        *args: t.Any,
        cfg: t.Dict[str, t.Any],
        rootpath: str = None,
        options: t.Dict[str, t.Any] = None,
        **kwargs: t.Any,
    ) -> None:
        """
        Initialize a Flasket application, and configure Gunicorn

        Parameters
        ----------
        cfg: dict

            dictionary containing configuration for the application

        rootpath: str

            location from where to load packages
        """
        self._flasket = GunicornFlasket(*args, gunicorn=self, cfg=cfg, rootpath=rootpath, options=options, **kwargs)

        # Rework the configuration
        cfg = self._flasket.config["server"]
        cfg["workers"] = self._get_cpu_workers(cfg["workers"])
        self.options = {
            "bind": f"{cfg['listen']}:{cfg['port']}",
            "pidfile": expandpath(cfg.get("pidfile")),
            "workers": cfg["workers"],
            "accesslog": "/dev/null",
            "errorlog": "/dev/null",
        }

        self._flasket.logger.info(f"Starting {cfg['workers']} workers...")
        super().__init__()

        # Use the Flask logger to log identically between flask and gunicorn
        # We have to specify a null FileStream handler, we'll remove it on first call
        # by using the _gunicorn identifier
        logger = logging.getLogger("gunicorn.error")
        default_handler._gunicorn = False
        logger.addHandler(default_handler)

    def init(self, parser: t.Any, opts: t.Any, args: t.Any) -> None:
        """
        Overloaded abstract method.
        """
        raise NotImplementedError

    def load_config(self) -> None:
        """
        Overloaded abstract method.

        Load Gunicorn configuration ourselves.
        """
        # Pass the configuration down to gunicorn.app.base.BaseApplication
        cfg = {k: v for k, v in self.options.items() if k in self.cfg.settings and v is not None}
        for k, v in cfg.items():
            self.cfg.set(k.lower(), v)

    def load(self) -> "Flasket":
        """
        Overloaded abstract method.

        Return a callable __call__ WSGI application
        """
        return self._flasket


@define(kw_only=True, slots=False)
class GunicornFlasket(Flasket):
    """
    Gunicorn specific Flasket
    """

    _gunicorn: GunicornApplication = field()

    def run(self) -> None:
        """
        Override run() since we don't take any parameters
        """
        self._gunicorn.run()
